#!/usr/bin/python2.5
#
# Copyright 2009 Roman Nurik
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines common geo math functions used throughout the library."""

import math

RADIUS = 6378135


def distance(p1, p2):
  """Calculates the great circle distance between two points using the
  law of cosines formula.
  
  Args:
    p1: A geotypes.Point or db.GeoPt indicating the first point.
    p2: A geotypes.Point or db.GeoPt indicating the second point.
  
  Returns:
    The 2D great-circle distance between the two given points, in meters.
  """
  p1lat, p1lon = math.radians(p1.lat), math.radians(p1.lon)
  p2lat, p2lon = math.radians(p2.lat), math.radians(p2.lon)
  return RADIUS * math.acos(math.sin(p1lat) * math.sin(p2lat) +
      math.cos(p1lat) * math.cos(p2lat) * math.cos(p2lon - p1lon))


def destination(p, heading, distance):
  """Approximates a destination point given a start point, an initial bearing,
  and a distance.
  
  Args:
    p: A geotypes.Point or db.GeoPt indicating the start point.
    heading: The initial bearing to follow, in degrees.
    distance: The distance to travel, in meters.
  
  Returns:
    A geotypes.Point approximately <distance> away from the starting point
    if following an initial bearing of <heading>.
  """
  plat = math.radians(p.lat)
  plon = math.radians(p.lon)
  
  angular_distance = distance * 1.0 / RADIUS
  heading = math.radians(heading)
  lat2 = math.asin(math.sin(plat) * math.cos(angular_distance) + 
                   math.cos(plat) * math.sin(angular_distance) *
                     math.cos(heading))
  
  return geotypes.Point(
      math.degrees(lat2),
      math.degrees(math.atan2(
          math.sin(heading) * math.sin(angular_distance) * math.cos(lat2),
          math.cos(angular_distance) - math.sin(plat) * math.sin(lat2))) + plon)


#!/usr/bin/python2.5
#
# Copyright 2009 Roman Nurik
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the GeoModel class for running basic geospatial queries on
single-point geographic entities in Google App Engine.

TODO(romannurik): document how bounding box and proximity queries work.
"""

__author__ = 'api.roman.public@gmail.com (Roman Nurik)'

import copy
import logging
import math
import sys
from google.appengine.ext import db
import geocell
import geomath
import geotypes
import util

DEBUG = False

DEFAULT_COST_FUNCTION = \
    lambda num_cells, resolution: 1e10000 if num_cells > \
                                  pow(geocell.GEOCELL_GRID_SIZE, 2) else 0


class GeoModel(db.Model):
  """A base model class for single-point geographically located entities.
  
  Attributes:
    location: A db.GeoPt that defines the single geographic point
        associated with this entity.
  """
  location = db.GeoPtProperty(required=True)
  
  # One for each of MAX_GEOCELL_RESOLUTION resolutions.
  #location_geocells = db.StringListProperty()
  location_geocell_1  = db.StringProperty()
  location_geocell_2  = db.StringProperty()
  location_geocell_3  = db.StringProperty()
  location_geocell_4  = db.StringProperty()
  location_geocell_5  = db.StringProperty()
  location_geocell_6  = db.StringProperty()
  location_geocell_7  = db.StringProperty()
  location_geocell_8  = db.StringProperty()
  location_geocell_9  = db.StringProperty()
  location_geocell_10 = db.StringProperty()
  location_geocell_11 = db.StringProperty()
  location_geocell_12 = db.StringProperty()
  location_geocell_13 = db.StringProperty()
  
  def update_location(self):
    """Updates the underlying geocell properties of the entity to match the
    entity's location property. A put() must occur after this call to save
    the changes to App Engine."""
    max_res_geocell = geocell.compute(self.location)
    for res in range(1, geocell.MAX_GEOCELL_RESOLUTION + 1):
      setattr(self, 'location_geocell_%d' % res, max_res_geocell[:res])

  @classmethod
  def bounding_box_fetch(cls, query, bbox, max_results=1000,
                         cost_function=None):
    """Fetches entities matching the given query with an additional filter
    matching only those entities that are inside of the given rectangular
    bounding box.
    
    Args:
      query: A db.Query on entities of this kind that should be additionally
          filtered by bounding box and subsequently fetched.
      bbox: A geotypes.Box indicating the bounding box to filter entities by.
      max_results: An optional int indicating the maximum number of desired
          results.
      cost_function: An optional function that accepts two arguments:
          * num_cells: the number of cells to search
          * resolution: the resolution of each cell to search
          and returns the `cost' of querying against this number of cells
          at the given resolution.
    
    Returns:
      The fetched entities.
    
    Raises:
      Any exceptions that .fetch() can raise.
    """
    # TODO(romannurik): Check for GqlQuery.
    results = []

    if cost_function is None:
      cost_function = DEFAULT_COST_FUNCTION
    query_geocells = geocell.best_bbox_search_cells(bbox, cost_function)
    
    if query_geocells:
      if query._Query__orderings:
        # NOTE(romannurik): since 'IN' queries seem boken in App Engine,
        # manually search each geocell and then merge results in-place
        cell_results = [copy.deepcopy(query)
            .filter('location_geocell_%d =' % len(search_cell),
                    search_cell)
            .fetch(max_results) for search_cell in query_geocells]
      
        # Manual in-memory sort on the query's defined ordering.
        query_orderings = query._Query__orderings or []
        def ordering_fn(ent1, ent2):
          for prop, direction in query_orderings:
            prop_cmp = cmp(getattr(ent1, prop), getattr(ent2, prop))
            if prop_cmp != 0:
              return prop_cmp if direction == 1 else -prop_cmp
        
          return -1  # Default ent1 < ent2.
      
        # Duplicates aren't possible so don't provide a dup_fn.
        util.merge_in_place(cmp_fn=ordering_fn, *cell_results)
        results = cell_results[0][:max_results]
      else:
        search_resolution = len(query_geocells[0])
        
        # NOTE: We can't pass in max_results because of non-uniformity of the
        # search.
        results = (query
            .filter('location_geocell_%d IN' % search_resolution,
                    query_geocells)
            .fetch(1000))
    else:
      results = []
    
    if DEBUG:
      logging.info('bbox query looked in %d geocells' % len(query_geocells))
    
    # In-memory filter.
    return [entity for entity in results if
        entity.location.lat >= bbox.south and
        entity.location.lat <= bbox.north and
        entity.location.lon >= bbox.west and
        entity.location.lon <= bbox.east]
  
  @classmethod
  def proximity_fetch(cls, query, center, max_results=10, max_distance=0):
    """Fetches at most <max_results> entities matching the given query,
    ordered by ascending distance from the given center point, and optionally
    limited by the given maximum distance.
    
    This method uses a greedy algorithm that starts by searching high-resolution
    geocells near the center point and gradually looking in lower and lower
    resolution cells until max_results entities have been found matching the
    given query and no closer possible entities can be found.
    
    Args:
      query: A db.Query on entities of this kind.
      center: A geotypes.Point or db.GeoPt indicating the center point around
          which to search for matching entities.
      max_results: An int indicating the maximum number of desired results.
          The default is 10, and the larger this number, the longer the fetch
          will take.
      max_distance: An optional number indicating the maximum distance to
          search, in meters.
    
    Returns:
      The fetched entities, sorted in ascending order by distance to the search
      center.
    
    Raises:
      Any exceptions that .fetch() can raise.
    """
    # TODO(romannurik): check for GqlQuery
    results = []
    
    searched_cells = set()

    # The current search geocell containing the lat,lon.
    cur_containing_geocell = geocell.compute(center)

    # The currently-being-searched geocells.
    # NOTES:
    #     * Start with max possible.
    #     * Must always be of the same resolution.
    #     * Must always form a rectangular region.
    #     * One of these must be equal to the cur_containing_geocell.
    cur_geocells = [cur_containing_geocell]
    
    closest_possible_next_result_dist = 0
    
    # Assumes both a and b are lists of (entity, dist) tuples, *sorted by dist*.
    # NOTE: This is an in-place merge, and there are guaranteed
    # no duplicates in the resulting list.
    def _merge_results_in_place(a, b):
      util.merge_in_place(a, b, cmp_fn=lambda x, y: cmp(x[1], y[1]),
                                dup_fn=lambda x, y: x[0].key() == y[0].key())
    
    sorted_edges = [(0,0)]
    sorted_edge_distances = [0]
    
    while cur_geocells:
      closest_possible_next_result_dist = sorted_edge_distances[0]
      if max_distance and closest_possible_next_result_dist > max_distance:
        break
      
      cur_geocells_unique = list(set(cur_geocells).difference(searched_cells))
      
      # Run query on the next set of geocells.
      cur_resolution = len(cur_geocells[0])
      temp_query = copy.deepcopy(query)  # TODO(romannurik): is this safe?
      temp_query.filter('location_geocell_%d IN' % cur_resolution,
                        cur_geocells_unique)
      
      # Update results and sort.
      new_results = temp_query.fetch(1000)
      if DEBUG:
        logging.info('fetch complete for %s' % (','.join(cur_geocells_unique),))
      
      searched_cells.update(cur_geocells)

      new_results = [(entity, geomath.distance(center, entity.location))
                     for entity in new_results]
      new_results = sorted(new_results, lambda dr1, dr2: cmp(dr1[1], dr2[1]))
      new_results = new_results[:max_results]
      
      # Merge new_results into results or the other way around, depending on
      # which is larger.
      if len(results) > len(new_results):
        _merge_results_in_place(results, new_results)
      else:
        _merge_results_in_place(new_results, results)
        results = new_results
      
      results = results[:max_results]
      
      sorted_edges, sorted_edge_distances = \
          util.distance_sorted_edges(cur_geocells, center)

      if len(results) == 0 or len(cur_geocells) == 4:
        # Either no results (in which case we optimize by not looking at
        # adjacents, go straight to the parent) or we've searched 4 adjacent
        # geocells, in which case we should now search the parents of those
        # geocells.
        cur_containing_geocell = cur_containing_geocell[:-1]
        cur_geocells = list(set([cell[:-1] for cell in cur_geocells]))
        if not cur_geocells or not cur_geocells[0]:
          break  # Done with search, we've searched everywhere.
      
      elif len(cur_geocells) == 1:
        # Get adjacent in one direction.
        # TODO(romannurik): Watch for +/- 90 degree latitude edge case geocells.
        nearest_edge = sorted_edges[0]
        cur_geocells.append(geocell.adjacent(cur_geocells[0], nearest_edge))
      
      elif len(cur_geocells) == 2:
        # Get adjacents in perpendicular direction.
        nearest_edge = util.distance_sorted_edges([cur_containing_geocell],
                                                  center)[0][0]
        if nearest_edge[0] == 0:
          # Was vertical, perpendicular is horizontal.
          perpendicular_nearest_edge = [x for x in sorted_edges if x[0] != 0][0]
        else:
          # Was horizontal, perpendicular is vertical.
          perpendicular_nearest_edge = [x for x in sorted_edges if x[0] == 0][0]
        
        cur_geocells.extend(
            [geocell.adjacent(cell, perpendicular_nearest_edge)
             for cell in cur_geocells])
      
      # We don't have enough items yet, keep searching.
      if len(results) < max_results:
        if DEBUG:
          logging.debug('have %d results but want %d results, '
                        'continuing search' % (len(results), max_results))
        continue
      
      if DEBUG:
        logging.debug('have %d results' % (len(results),))
      
      # If the currently max_results'th closest item is closer than any
      # of the next test geocells, we're done searching.
      current_farthest_returnable_result_dist = \
          geomath.distance(center, results[max_results - 1][0].location)
      if (closest_possible_next_result_dist >=
          current_farthest_returnable_result_dist):
        if DEBUG:
          logging.debug('DONE next result at least %f away, '
                        'current farthest is %f dist' %
                        (closest_possible_next_result_dist,
                         current_farthest_returnable_result_dist))
        break
      
      if DEBUG:
        logging.debug('next result at least %f away, '
                      'current farthest is %f dist' %
                      (closest_possible_next_result_dist,
                       current_farthest_returnable_result_dist))
    
    if DEBUG:
      logging.info('proximity query looked '
                   'in %d geocells' % len(searched_cells))
    
    return [entity for (entity, dist) in results[:max_results]
                   if not max_distance or dist < max_distance]


#!/usr/bin/python2.5
#
# Copyright 2009 Roman Nurik
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines some useful geo types such as points and boxes."""

__author__ = 'api.roman.public@gmail.com (Roman Nurik)'


class Point(object):
  """A two-dimensional point in the [-90,90] x [-180,180] latitude/longitude
  space.
  
  Attributes:
    lat: A float in the range [-90,90] indicating the point's latitude.
    lon: A float in the range [-180,180] indicating the point's longitude.
  """
  
  lat = None
  lon = None
  
  def __init__(self, lat, lon):
    """Initializes a point with the given latitude and longitude."""
    if -90 > lat or lat > 90:
      raise ValueError("Latitude must be in [-90, 90] but was %f" % lat)
    if -180 > lon or lon > 180:
      raise ValueError("Longitude must be in [-180, 180] but was %f" % lon)
      
    self.lat = lat
    self.lon = lon
  
  def __str__(self):
    return '(%f, %f)' % (self.lat, self.lon)


class Box(object):
  """A two-dimensional rectangular region defined by Northeast and Southwest
  points.
  
  Attributes:
    north_east: A read-only geotypes.Point indicating the box's Northeast
        coordinate.
    south_west: A read-only geotypes.Point indicating the box's Southwest
        coordinate.
    north: A float indicating the box's North latitude.
    east: A float indicating the box's East longitude.
    south: A float indicating the box's South latitude.
    west: A float indicating the box's West longitude.
  """
  
  _ne = None
  _sw = None
  
  def __init__(self, north, east, south, west):
    # TODO(romannurik): swap south/north if necessary
    # TODO(romannurik): port geojs LatLonBox here
    self._ne = Point(north, east)
    self._sw = Point(south, west)
  
  north_east = property(lambda self: self._ne)
  south_west = property(lambda self: self._sw)
  
  def _set_north(self, val):
    self._ne.lat = val
  north = property(lambda self: self._ne.lat, _set_north)
  
  def _set_east(self, val):
    self._ne.lat = val
  east  = property(lambda self: self._ne.lon, _set_east)
  
  def _set_south(self, val):
    self._sw.lat = val
  south = property(lambda self: self._sw.lat, _set_south)
  
  def _set_west(self, val):
    self._sw.lon = val
  west  = property(lambda self: self._sw.lon, _set_west)
  
  def __str__(self):
    return '(N:%f, E:%f, S:%f, W:%f)' % (self.north, self.east,
                                         self.south, self.west)


#!/usr/bin/python2.5
#
# Copyright 2009 Roman Nurik
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the notion of `geocells' and exposes methods to operate on them.

A geocell is a hexadecimal string that defines a two dimensional rectangular
region inside the [-90,90] x [-180,180] latitude/longitude space. A geocell's
`resolution' is its length. For most practical purposes, at high resolutions,
geocells can be treated as single points.

Much like geohashes (see http://en.wikipedia.org/wiki/Geohash), geocells are
hierarchical, in that any prefix of a geocell is considered its ancestor, with
geocell[:-1] being geocell's immediate parent cell.

To calculate the rectangle of a given geocell string, first divide the
[-90,90] x [-180,180] latitude/longitude space evenly into a 4x4 grid like so:

             +---+---+---+---+ (90, 180)
             | a | b | e | f |
             +---+---+---+---+
             | 8 | 9 | c | d |
             +---+---+---+---+
             | 2 | 3 | 6 | 7 |
             +---+---+---+---+
             | 0 | 1 | 4 | 5 |
  (-90,-180) +---+---+---+---+

NOTE: The point (0, 0) is at the intersection of grid cells 3, 6, 9 and c. And,
      for example, cell 7 should be the sub-rectangle from
      (-45, 90) to (0, 180).

Calculate the sub-rectangle for the first character of the geocell string and
re-divide this sub-rectangle into another 4x4 grid. For example, if the geocell
string is `78a', we will re-divide the sub-rectangle like so:

               .                   .
               .                   .
           . . +----+----+----+----+ (0, 180)
               | 7a | 7b | 7e | 7f |
               +----+----+----+----+
               | 78 | 79 | 7c | 7d |
               +----+----+----+----+
               | 72 | 73 | 76 | 77 |
               +----+----+----+----+
               | 70 | 71 | 74 | 75 |
  . . (-45,90) +----+----+----+----+
               .                   .
               .                   .

Continue to re-divide into sub-rectangles and 4x4 grids until the entire
geocell string has been exhausted. The final sub-rectangle is the rectangular
region for the geocell.

TODO(romannurik): Unit tests!
"""

__author__ = 'api.roman.public@gmail.com (Roman Nurik)'

import os.path
import sys
import geotypes
import geomath

MAX_GEOCELL_RESOLUTION = 13  # The maximum *practical* geocell resolution.
GEOCELL_GRID_SIZE = 4
GEOCELL_ALPHABET = '0123456789abcdef'

#def _best_radius_search_cells(center, distance, cost_function):
#  n = geomath.destination(center, 0, distance).lat
#  e = geomath.destination(center, 90, distance).lon
#  s = geomath.destination(center, 180, distance).lat
#  w = geomath.destination(center, 270, distance).lon
#  
#  n, s = max(n, s), min(n, s)
#  e, w = max(e, w), min(e, w)
#  
#  return best_bbox_search_cells((n, e, s, w), cost_function)

def best_bbox_search_cells(bbox, cost_function):
  """Returns the most efficient set of geocells to search in a bounding box
  query, given a cost function.
  
  This method is guaranteed to return a set of geocells having the same
  resolution.
  
  Args:
    bbox: A geotypes.Box indicating the bounding box being searched.
    cost_function: A function that accepts two arguments:
        * num_cells: the number of cells to search
        * resolution: the resolution of each cell to search
        and returns the `cost' of querying against this number of cells
        at the given resolution.
  
  Returns:
    A list of geocell strings that contain the given box.
  """
  cell_ne = compute(bbox.north_east, resolution=MAX_GEOCELL_RESOLUTION)
  cell_sw = compute(bbox.south_west, resolution=MAX_GEOCELL_RESOLUTION)
  
  min_cost = 1e10000  # Practical infinity.
  min_cost_cell_set = None
  
  # First find the common prefix, if there is one.. this will be the base
  # resolution.. i.e. we don't have to look at any higher resolution cells.
  min_resolution = len(os.path.commonprefix([cell_sw, cell_ne]))
  
  # Iteravely calculate all possible sets of cells that wholely contain
  # the requested bounding box.
  for cur_resolution in range(min_resolution, MAX_GEOCELL_RESOLUTION + 1):
    cur_ne = cell_ne[:cur_resolution]
    cur_sw = cell_sw[:cur_resolution]
    
    num_cells = interpolation_count(cur_ne, cur_sw)
    if num_cells > 300:
      continue
    
    cell_set = sorted(interpolate(cur_ne, cur_sw))
    simplified_cells = []
    
    '''
    # NOTE: this may be moot
    # try to simplify the geocells by prefix matching
    consecs = {}
    for i in range(1, len(cell_set)):
      cp = commonprefix([cell_set[i], cell_set[i - 1]])
      # increment count for this prefix and all parent prefixes
      for s in range(len(cp)):
        pfx = cp[:s + 1]
        if not pfx in consecs:
          consecs[pfx] = 0
        elif consecs[pfx] < 0: # this prefix is already marked to be collapsed
          continue
        else:
          consecs[pfx] += 1
          if consecs[pfx] >= pow(16, cur_resolution - len(pfx)):
            consecs[pfx] = -1 # mark the prefix to be collapsed
            simplified_cells.append(pfx)
    
    print >> sys.stderr, simplified_cells
    
    for sc in sorted(simplified_cells):
      cell_set = filter(lambda c: not c.startswith(sc), cell_set)
      simplified_cells = filter(lambda c: c == sc or not c.startswith(sc),
                                simplified_cells)
    cell_set.extend(simplified_cells)
    '''
    
    cost = cost_function(num_cells=len(cell_set), resolution=cur_resolution) 
    
    # TODO(romannurik): See if this resolution is even possible, as in the 
    # future cells at certain resolutions may not be stored.
    if cost <= min_cost:
      min_cost = cost
      min_cost_cell_set = cell_set
      #print >> sys.stderr, 'boxes=%d, cost=%f' % (len(cell_set), cost)
    else:
      # Once the cost starts rising, we won't be able to do better, so abort.
      break
  
  return min_cost_cell_set


def collinear(cell1, cell2, direction=0):
  """Determines whether the given cells are collinear in the given
  dimension--that is, whether or not they are in the same row (direction=0)
  or in the same column (direction=1).
  
  Args:
    cell1: The first geocell string.
    cell2: The second geocell string.
    direction: An int, where 0 invokes a horizontal collinearity test and 1
        invokes a vertical collinearity test.
  
  Returns:
    A bool indicating whether or not the given cells are collinear in the given
    dimension.
  """
  for i in range(min(len(cell1), len(cell2))):
    x1, y1 = _subdiv_xy(cell1[i])
    x2, y2 = _subdiv_xy(cell2[i])
    
    # Check horizontal collinearity (assure y's are always the same).
    if direction == 0 and y1 != y2:
      return False
    
    # Check vertical collinearity (assure x's are always the same).
    if direction == 1 and x1 != x2:
      return False
  
  return True


def interpolate(cell_ne, cell_sw):
  """Generates the set of cells in the grid created by interpolating from the
  given Northeast geocell to the given Southwest geocell.
  
  Assumes the Northeast geocell is actually Northeast of Southwest geocell.
  
  Arguments:
    cell_ne: The Northeast geocell string.
    cell_sw: The Southwest geocell string.
  
  Returns:
    A list of geocell strings in the interpolation.
  """
  # 2D array, will later be flattened.
  cell_set = [[cell_sw]]
  
  # First get adjacent geocells across until Southeast--collinearity with
  # Northeast in vertical direction (0) means we're at Southeast.
  while not collinear(cell_set[0][-1], cell_ne, 1):
    cell_tmp = adjacent(cell_set[0][-1], (1, 0))
    if cell_tmp is None:
      break
    cell_set[0].append(cell_tmp)
  
  # Then get adjacent geocells upwards.
  while cell_set[-1][-1] != cell_ne:
    cell_tmp_row = [adjacent(g, (0, 1)) for g in cell_set[-1]]
    if cell_tmp_row[0] is None:
      break
    cell_set.append(cell_tmp_row)
  
  # Flatten cell_set, since it's currently a 2D array.
  cell_set = [g for inner in cell_set for g in inner]
  return cell_set


def interpolation_count(cell_ne, cell_sw):
  """Computes the number of cells in the grid created by interpolating from the
  given Northeast geocell to the given Southwest geocell.
  
  Assumes the Northeast geocell is actually Northeast of Southwest geocell.
  
  Arguments:
    cell_ne: The Northeast geocell string.
    cell_sw: The Southwest geocell string.
  
  Returns:
    An int, indicating the number of geocells in the interpolation.
  """
  bbox_ne = compute_box(cell_ne)
  bbox_sw = compute_box(cell_sw)
  
  return int(((bbox_ne.east - bbox_sw.west) / (bbox_sw.east - bbox_sw.west)) * \
            ((bbox_ne.north - bbox_sw.south) / (bbox_sw.north - bbox_sw.south)))


def all_adjacents(cell):
  """Calculates all of the given geocell's adjacent geocells, including None
  values where there is no neighboring cell in a possible direction.
  
  Args:
    cell: The geocell string for which to calculate adjacent/neighboring cells.
  
  Returns:
    A list of 8 geocell strings and/or None values indicating adjacent cells.
  """
  return [adjacent(cell, d) for d in [(-1,-1), ( 0,-1), ( 1,-1), (-1, 0),
                                      ( 1, 0), (-1, 1), ( 0, 1), ( 1, 1)]]


def adjacent(cell, dir):
  """Calculates the geocell adjacent to the given cell in the given direction.
  
  Args:
    cell: The geocell string whose neighbor is being calculated.
    dir: A (x, y) tuple indicating direction, where x and y can be -1, 0, or 1.
        -1 corresponds to West for x and South for y, and
         1 corresponds to East for x and North for y.
  
  Returns:
    The geocell adjacent to the given cell in the given direction, or None if
    there is no such cell.
  """
  if cell is None:
    return None
  
  dx = dir[0]
  dy = dir[1]
  
  cell_adj_arr = list(cell)  # Split the geocell string characters into a list.
  i = len(cell_adj_arr) - 1
  
  while i >= 0 and (dx != 0 or dy != 0):
    x, y = _subdiv_xy(cell_adj_arr[i])
    
    # Horizontal adjacency.
    if dx == -1:  # Asking for left.
      if x == 0:  # At left of parent cell.
        x = GEOCELL_GRID_SIZE - 1  # Becomes right edge of adjacent parent.
      else:
        x -= 1  # Adjacent, same parent.
        dx = 0  # Done with x.
    elif dx == 1:  # Asking for right.
      if x == GEOCELL_GRID_SIZE - 1:  # At right of parent cell.
        x = 0  # Becomes left edge of adjacent parent.
      else:
        x += 1  # Adjacent, same parent.
        dx = 0  # Done with x.
    
    # Vertical adjacency.
    if dy == 1:  # Asking for above.
      if y == GEOCELL_GRID_SIZE - 1:  # At top of parent cell.
        y = 0  # Becomes bottom edge of adjacent parent.
      else:
        y += 1  # Adjacent, same parent.
        dy = 0  # Done with y.
    elif dy == -1:  # Asking for below.
      if y == 0:  # At bottom of parent cell.
        y = GEOCELL_GRID_SIZE - 1  # Becomes top edge of adjacent parent.
      else:
        y -= 1  # Adjacent, same parent.
        dy = 0  # Done with y.
    
    cell_adj_arr[i] = _subdiv_char((x,y))
    i -= 1
  
  # If we're not done with y then it's trying to wrap vertically,
  # which is a failure.
  if dy != 0:
    return None
  
  # At this point, horizontal wrapping is done inherently.
  return ''.join(cell_adj_arr)


def contains_point(cell, point):
  """Returns whether or not the given cell contains the given point."""
  return compute(point, len(cell)) == cell


def point_distance(cell, point):
  """Returns the shortest great circle distance from a given point to the given
  geocell's rectangle.

  If the point is inside the cell, the shortest distance is always to a `edge'
  of the cell rectangle. If the point is outside the cell, the shortest distance
  will be to either a `edge' or `corner' of the cell rectangle.

  Returns:
    The shortest distance from the point to the geocell's rectangle, in meters.
  """
  bbox = compute_box(cell)
  
  between_w_e = bbox.west <= point.lon and point.lon <= bbox.east
  between_n_s = bbox.south <= point.lat and point.lat <= bbox.north
  
  if between_w_e:
    if between_n_s:
      # Inside the geocell.
      return min(geomath.distance(point, (bbox.south, point.lon)),
                 geomath.distance(point, (bbox.north, point.lon)),
                 geomath.distance(point, (point.lat, bbox.east)),
                 geomath.distance(point, (point.lat, bbox.west)))
    else:
      return min(geomath.distance(point, (bbox.south, point.lon)),
                 geomath.distance(point, (bbox.north, point.lon)))
  else:
    if between_n_s:
      return min(geomath.distance(point, (point.lat, bbox.east)),
                 geomath.distance(point, (point.lat, bbox.west)))
    else:
      # TODO(romannurik): optimize
      return min(geomath.distance(point, (bbox.south, bbox.east)),
                 geomath.distance(point, (bbox.north, bbox.east)),
                 geomath.distance(point, (bbox.south, bbox.west)),
                 geomath.distance(point, (bbox.north, bbox.west)))


def compute(point, resolution=MAX_GEOCELL_RESOLUTION):
  """Computes the geocell string containing the given point to the given
  resolution.
  
  This is a simple 16-tree lookup to an arbitrary depth (resolution).
  
  Args:
    point: The geotypes.Point to compute the cell for.
    resolution: An int indicating the resolution of the cell to compute.
  
  Returns:
    The geocell string containing the given point, of length <resolution>.
  """
  north = 90.0
  south = -90.0
  east = 180.0
  west = -180.0
  
  cell = ''
  while len(cell) < resolution:
    subcell_lon_span = (east - west) / GEOCELL_GRID_SIZE
    subcell_lat_span = (north - south) / GEOCELL_GRID_SIZE
    
    x = min(int(GEOCELL_GRID_SIZE * (point.lon - west) / (east - west)),
            GEOCELL_GRID_SIZE - 1)
    y = min(int(GEOCELL_GRID_SIZE * (point.lat - south) / (north - south)),
            GEOCELL_GRID_SIZE - 1)
    
    cell += _subdiv_char((x,y))
    
    south += subcell_lat_span * y
    north = south + subcell_lat_span
    
    west += subcell_lon_span * x
    east = west + subcell_lon_span
  
  return cell


def compute_box(cell):
  """Computes the rectangular boundaries (bounding box) of the given geocell.
  
  Args:
    cell: The geocell string whose boundaries are to be computed.
  
  Returns:
    A geotypes.Box corresponding to the rectangular boundaries of the geocell.
  """
  if cell is None:
    return None

  bbox = geotypes.Box(90.0, 180.0, -90.0, -180.0)
  
  while len(cell) > 0:
    subcell_lon_span = (bbox.east - bbox.west) / GEOCELL_GRID_SIZE
    subcell_lat_span = (bbox.north - bbox.south) / GEOCELL_GRID_SIZE
    
    x, y = _subdiv_xy(cell[0])
    
    bbox = geotypes.Box(bbox.south + subcell_lat_span * (y + 1),
                        bbox.west  + subcell_lon_span * (x + 1),
                        bbox.south + subcell_lat_span * y,
                        bbox.west  + subcell_lon_span * x)
    
    cell = cell[1:]
  
  return bbox


def is_valid(cell):
  """Returns whether or not the given geocell string defines a valid geocell."""
  return bool(cell and reduce(lambda val, c: val and c in GEOCELL_ALPHABET,
                              cell, True))


def children(cell):
  """Calculates the immediate children of the given geocell.
  
  For example, the immediate children of `a' are `a0', `a1', ..., `af'.
  """
  return [cell + chr for chr in GEOCELL_ALPHABET]


def _subdiv_xy(c):
  """Returns the (x, y) position of the given geocell character
  in the 4x4 alphabet grid."""
  # NOTE: This only works for grid size 4.
  c = GEOCELL_ALPHABET.index(c)
  return ((c & 4) >> 1 | (c & 1) >> 0,
          (c & 8) >> 2 | (c & 2) >> 1)


def _subdiv_char(pos):
  """Returns the geocell character in the 4x4 alphabet grid at pos. (x, y)."""
  # NOTE: This only works for grid size 4.
  return GEOCELL_ALPHABET[
      (pos[1] & 2) << 2 |
      (pos[0] & 2) << 1 |
      (pos[1] & 1) << 1 |
      (pos[0] & 1) << 0]


if __name__ == '__main__':
  pass


#!/usr/bin/python2.5
#
# Copyright 2009 Roman Nurik
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines utility functions used throughout the geocell/GeoModel library."""

__author__ = 'api.roman.public@gmail.com (Roman Nurik)'

import geocell
import geomath
import geotypes

def merge_in_place(*lists, **kwargs):
  """Merges an arbitrary number of pre-sorted lists in-place, into the first
  list, possibly pruning out duplicates.
  
  Args:
    list1: The first, sorted list into which the other lists should be merged.
    list2: A subsequent, sorted list to merge into the first.
    ...
    listn:  "   "
    cmp_fn: An optional binary comparison function that compares objects across
        lists and determines the merged list's sort order.
    dup_fn: An optional binary comparison function that should return True if
        the given objects are equivalent and one of them can be pruned from the
        resulting merged list.
  
  Returns:
    list1, in-placed merged wit the other lists, or an empty list if no lists
    were specified.
  """
  cmp_fn = kwargs.get('cmp_fn') or cmp
  dup_fn = kwargs.get('dup_fn') or None
  
  if not lists:
    return []
  
  reverse_indices = [len(arr) for arr in lists]
  aggregate_reverse_index = sum(reverse_indices)
  
  while aggregate_reverse_index > 0:
    pull_arr_index = None
    pull_val = None
    
    for i in range(len(lists)):
      if reverse_indices[i] == 0:
        # Reached the end of this list.
        continue
      
      if (pull_arr_index is not None and
          dup_fn and dup_fn(lists[i][-reverse_indices[i]], pull_val)):
        # Found a duplicate, advance the index of the list in which the
        # duplicate was found.
        reverse_indices[i] -= 1
        aggregate_reverse_index -= 1
        continue
      
      if (pull_arr_index is None or
          cmp_fn(lists[i][-reverse_indices[i]], pull_val) < 0):
        # Found a lower value.
        pull_arr_index = i
        pull_val = lists[i][-reverse_indices[i]]
    
    if pull_arr_index != 0:
      # Add the lowest found value in place into the first array.
      lists[0].insert(len(lists[0]) - reverse_indices[0], pull_val)
    
    aggregate_reverse_index -= 1
    reverse_indices[pull_arr_index] -= 1
  
  return lists[0]


def distance_sorted_edges(cells, point):
  """Returns the edges of the rectangular region containing all of the
  given geocells, sorted by distance from the given point, along with
  the actual distances from the point to these edges.
  
  Args:
    cells: The cells (should be adjacent) defining the rectangular region
        whose edge distances are requested.
    point: The point that should determine the edge sort order.
  
  Returns:
    A list of (direction, distance) tuples, where direction is the edge
    and distance is the distance from the point to that edge. A direction
    value of (0,-1), for example, corresponds to the South edge of the
    rectangular region containing all of the given geocells.
  """
  # TODO(romannurik): Assert that lat,lon are actually inside the geocell.
  boxes = [geocell.compute_box(cell) for cell in cells]
  
  max_box = geotypes.Box(max([box.north for box in boxes]),
                         max([box.east for box in boxes]),
                         min([box.south for box in boxes]),
                         min([box.west for box in boxes]))
  return zip(*sorted([
      ((0,-1), geomath.distance(geotypes.Point(max_box.south, point.lon),
                                point)),
      ((0,1),  geomath.distance(geotypes.Point(max_box.north, point.lon),
                                point)),
      ((-1,0), geomath.distance(geotypes.Point(point.lat, max_box.west),
                                point)),
      ((1,0),  geomath.distance(geotypes.Point(point.lat, max_box.east),
                                point))],
      lambda x, y: cmp(x[1], y[1])))


"""Setup script."""

import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(*rnames)).read()


setup (
    name='geomodel',
    version='0.1.0',
    author="Roman Nurik",
    author_email="api dot roman dot public at gmail dot com (Roman Nurik)",
    description="Indexing and querying geospatial data in App Engine.",
    long_description=(
        read('README.txt')
        + '\n\n' +
        read('src', 'geo', 'tests', 'geomodel.txt')
        ),
    license="Apache License 2.0",
    keywords="appengine gae geomodel google maps",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        ],
    url='http://code.google.com/p/geomodel/',
    packages=find_packages('src'),
    include_package_data=True,
    package_dir={'': 'src'},
    install_requires=[
        'setuptools',
        ],
    zip_safe=False,
    )
